#ifndef __READCPT_H__

#define __READCPT_H__

#include <vl/VLd.h>
#include <vl/VLf.h>
#include <limits>

class BoundingBox3d
{
public:
	typedef vl::Vec3f Vec3;
	// initialize to be "empty"
	BoundingBox3d()
		: min_(vl::vl_0), max_(vl::vl_0)
	{
	}

	BoundingBox3d( Vec3 min, Vec3 max )
		: min_(min), max_(max)
	{
	}

	Vec3 minimum() const { return min_; }
	Vec3 maximum() const { return max_; }

private:
	Vec3 min_;
	Vec3 max_;
};

class DistanceFieldException
{
public:
	DistanceFieldException( std::string msg )
		: msg_(msg)
	{
	}

	std::string message() { return msg_; }

private:
	std::string msg_;
};

class SignedDistanceRepresentation
{
public:
	// Use single precision for distance transform.  This
	// is okay because error does not accumulate (as it would
	// in a solver, for example).
	typedef float Real;
	typedef vl::Vec3f Vec3;

	// read the distance, gradient fields in from the specified file
	SignedDistanceRepresentation(const std::string distanceFilename,
		const std::string gradientFilename);

	template <typename InVec3>
	bool outside( const InVec3& pos ) const
	{
		return (
			pos[0] >= (_bounds.maximum())[0] ||
			pos[1] >= (_bounds.maximum())[1] ||
			pos[2] >= (_bounds.maximum())[2] ||
			pos[0] <= (_bounds.minimum())[0] ||
			pos[1] <= (_bounds.minimum())[1] ||
			pos[2] <= (_bounds.minimum())[2]);
	}

	template <typename InVec3>
	Real distance( const InVec3& position ) const
	{
		if( outside( position ) )
			return std::numeric_limits<float>::max();				// Too big to matter, anyway

		return interpolateQuantity( _distance, position, 0.0f );
	}

	template <typename InVec3>
		vl::Vec3d gradient( const InVec3& position ) const
	{
		// Not sure what to do here; return a 0 gradient
		// will probably break other code since it tends to assume
		// gradient has magnitude 1 (this is the Eikonal equation,
		// after all).
		assert( !outside(position) );

		Vec3 res = interpolateQuantity( _gradientDistance, position, Vec3(vl::vl_0) );

		return vl::Vec3d( res[0], res[1], res[2] );
	}

	BoundingBox3d bounds() const { return _bounds; }

private:
	template <typename T>
	const T& access( const std::vector<T>& array, const int i0, const int i1, const int i2 ) const
	{
		return array[ i0 * _strides[0] + i1 * _strides[1] + i2 * _strides[2] ];
	}

	template <typename T, typename InVec3>
	T interpolateQuantity( const std::vector<T>& vec,
		InVec3 pos, T quantity ) const
	{
		// precondition: pos must be contained within the bounds.
		assert( !outside( pos ) );

		// We are performing trilinear interpolation here
		Vec3 indicesReal = Vec3(pos[0], pos[1], pos[2]) - _bounds.minimum();
		indicesReal[0] /= _delta[0];
		indicesReal[1] /= _delta[1];
		indicesReal[2] /= _delta[2];

		int indices[3] = {
			static_cast<int>(indicesReal[0]),
			static_cast<int>(indicesReal[1]),
			static_cast<int>(indicesReal[2]) };

		Vec3 offsets(
			indicesReal[0] - static_cast<Real>(indices[0]),
			indicesReal[1] - static_cast<Real>(indices[1]),
			indicesReal[2] - static_cast<Real>(indices[2]) );

		quantity += (1.0f-offsets[0])*(1.0f-offsets[1])*(1.0f-offsets[2])*access(vec, indices[0], indices[1], indices[2]);
		quantity += (offsets[0])*(1.0f-offsets[1])*(1.0f-offsets[2])*access(vec, indices[0] + 1, indices[1], indices[2]);
		quantity += (1.0f-offsets[0])*(offsets[1])*(1.0f-offsets[2])*access(vec, indices[0], indices[1] + 1, indices[2]);
		quantity += (1.0f-offsets[0])*(1.0f-offsets[1])*(offsets[2])*access(vec, indices[0], indices[1], indices[2] + 1);
		quantity += (offsets[0])*(1.0f-offsets[1])*(offsets[2])*access(vec, indices[0] + 1, indices[1], indices[2] + 1);
		quantity += (1.0f-offsets[0])*(offsets[1])*(offsets[2])*access(vec, indices[0], indices[1] + 1, indices[2] + 1);
		quantity += (offsets[0])*(offsets[1])*(1.0f-offsets[2])*access(vec, indices[0] + 1, indices[1] + 1, indices[2]);
		quantity += (offsets[0])*(offsets[1])*(offsets[2])*access(vec, indices[0] + 1, indices[1] + 1, indices[2] + 1);
		return quantity;
	}

private:
	std::vector<Real> _distance;
	std::vector<Vec3> _gradientDistance;
	size_t _gridDimensions[3];
	int _strides[3];
	BoundingBox3d _bounds;
	Vec3 _delta;
};

#endif
